--
-- SafeTricks.lua: Safe ToTT
-- Yawning (Crushridge-US)
--

--
-- Create our addon.
--

SafeTricks = LibStub("AceAddon-3.0"):NewAddon("SafeTricks", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0");
local AceConfig = LibStub("AceConfig-3.0");
local AceConfigDialog = LibStub("AceConfigDialog-3.0");

--
-- Version
--

SafeTricks.version = 20100421;

--
-- Debugging spew control.
--

function SafeTricks:DBGF(...)
	if (self.debug == true) then
		self:Print(...)
	end
end

--
-- Helper to remove a buff.
--

function SafeTricks:RemoveBuff(id)
	local i = 1
	local n,_,_,_,_,_,_,_,_,_,s= UnitBuff("player",i);

	while (n ~= nil) do
		if (s == id) then
			CancelUnitBuff("player", i);
			return true
		end
		i = i + 1;
		n,_,_,_,_,_,_,_,_,_,s= UnitBuff("player",i);
	end

	return false;
end

--
-- Roster maintenance.
-- Since there is a potential for PARTY_MEMBERS_CHANGED to fire rather frequently, throttle how often
-- we iterate over the player's group by using a timer.
--

function SafeTricks:ScheduleRosterRescan()
	self:DBGF("ScheduleRosterRescan()");

	self:CancelTimer(self.rosterTimer, true);
	self.rosterTimer = self:ScheduleTimer("DoRosterUpdate", 0.2);  -- 200 ms should be frequent enough
end

function SafeTricks:IsUnitTank(unitId, isRaid)
	local rval = false;

	if (isRaid == true) then
		if (GetPartyAssignment("MAINTANK", unitId) ~= nil) then
			rval = true;
		else
			rval = false;
		end
	else
		rval = UnitGroupRolesAssigned(unitId);
	end

	return rval;
end

function SafeTricks:DoRosterUpdate()
	self:DBGF("DoRosterUpdate()");

	if (self.isInBG == true) then
		self:DBGF("Not doing anything since we are in a BG at the moment.");
		return;
	end

	--
	-- Roster entrys are like thus:
	-- self.roster[name] = {
	-- 	guid = guid,
	-- 	isTank = bool,
	-- 	isRogue = bool,
	-- 	isAlwaysDPS = bool,
	-- 	isAlwaysTank = bool,
	-- }
	--

	wipe(self.roster);
	wipe(self.guidRoster);

	local prefix = nil;
	local maxMembers = 0;
	local isRaid = false;

	local unit = "";
	local name = "";
	local guid = "";
	local isTank = false;
	local isRogue = false;
	local isAlwaysTank = false;
	local isAlwaysDPS = false;

	if (GetNumRaidMembers() > 0) then
		-- Raid Group.
		prefix = "raid";
		maxMembers = GetNumRaidMembers();
		isRaid = true;
	elseif (GetNumPartyMembers() > 0) then
		-- 5 man group.
		prefix = "party";
		maxMembers = GetNumPartyMembers();
	end

	if (maxMembers > 0) then
		for i = 1, maxMembers do
			unit = prefix..i;
			name = GetUnitName(unit, true);
			guid = UnitGUID(unit);

			isTank = SafeTricks:IsUnitTank(unit, isRaid);

			if ((select(2,UnitClass(unit)) == "ROGUE")) then
				isRogue = true;
			else
				isRogue = false;
			end

			self.roster[name] = {
				guid = guid,
				isTank = isTank,
				isRogue = isRogue,
				isAlwaysDPS = isAlwaysDPS,
				isAlwaysTank = isAlwaysTank
			};
		end

		self:DoGUIDRosterUpdate();
	end
end

function SafeTricks:DoGUIDRosterUpdate()
	self.applyLogic = false;

	wipe(self.guidRoster);

	for k, v in pairs(self.db.char.alwaysTank) do
		if (self.roster[v] ~= nil) then
			self.roster[v].isAlwaysTank = true;
			self:DBGF("Tank override "..v);
		end
	end

	for k, v in pairs(self.db.char.alwaysDPS) do
		if (self.roster[v] ~= nil) then
			self.roster[v].isAlwaysDPS = true;
			self:DBGF("DPS override "..v);
		end
	end

	--
	-- Mode 0: +Threat Tanks, +DMG DPS, (+Threat All on 0 Tanks)
	-- Mode 1: +DMG Tanks, +DMG DPS (+DMG All on 0 Tanks)
	-- Mode 2: +Threat Tanks, +DMG DPS, +Threat Rogues (+Threat All on 0 Tanks, even if there are Rogues.)
	--
	for k, v in pairs(self.roster) do
		if (self.db.char.mode == 0) then
			if ((v.isAlwaysTank == true) or ((v.isTank == true) and (v.isAlwaysDPS == false))) then
				self.guidRoster[v.guid] = k;
				self.applyLogic = true;
			end
		elseif (self.db.char.mode == 1) then
			self.applyLogic = true;
			if (v.isAlwaysTank == true) then
				self.guidRoster[v.guid] = k;
			end
		elseif (self.db.char.mode == 2) then
			if ((v.isAlwaysTank == true) or ((v.isTank == true) and (v.isAlwaysDPS == false))) then
				self.guidRoster[v.guid] = k;
				self.applyLogic = true;
			elseif ((v.isRogue == true) and (v.isAlwaysDPS == false)) then
				self.guidRoster[v.guid] = k;
			end
		end
	end
end

-- 
-- Check to see if we are in a BG, and  update our status.
--

function SafeTricks:CheckForBG()
	local isInstance, instanceType = IsInInstance();

	if ((isInstance == 1) and (instanceType == "pvp")) then
		self.isInBG = true;
	else
		self.isInBG = false;
	end
end

--
-- Send an announcement to the appropriate sink.
--

function SafeTricks:SendTricksMsg(msg)
	if (self.isInBG == true) then
		self:DBGF("Not sending anything since we are in a BG at the moment.");
		return;
	end

	if (self.isMuted == true) then
		return;
	end

	if (self.db.char.partyAnnounce == true) then
		if (GetNumPartyMembers() > 0) then
			SendChatMessage(msg, "PARTY");
		end
	end

	if (self.db.char.raidAnnounce == true) then
		if (GetNumRaidMembers() > 0) then
			SendChatMessage(msg, "RAID");
		end
	end

	if (self.db.char.channelAnnounce == true) then
		if (self.db.char.channelSink ~= nil) then
			local id = GetChannelName(self.db.char.channelSink);
			if (id > 0) then
				SendChatMessage(msg, "CHANNEL", nil, id);
			end
		end
	end
end

--
-- COMBAT_LOG_EVENT_UNFILTERED
--

function SafeTricks:COMBAT_LOG_EVENT_UNFILTERED(ev, ...)
	local timestamp, combatEvent, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags;
	local spellID, spellName, spellSchool, auraType;

	-- No need to do anything if we are in a BG.
	if (self.isInBG == true) then
		return;
	end

	combatEvent = select(2, ...);
	srcGUID = select(3, ...);
	dstGUID = select(6, ...);
	dstName = select(7, ...);

	if (srcGUID == UnitGUID("player")) then
		spellID = select(9, ...);

		if (combatEvent == "SPELL_CAST_SUCCESS") then
			if (spellID == 57934) then
				self:DBGF("ToTT: "..dstName);
				self.curTarget = dstName;

				if (self.applyLogic == true) then
					self.shouldRemove = true;

					if (self.guidRoster[dstGUID] ~= nil) then
						self:DBGF("Tank, doing nothing.");
						self.shouldRemove = false;
					elseif (self.isForced == true) then
						self:DBGF("Forced threat xfer.");
						self.isForced = false;
						self.shouldRemove = false;
					else
						self:DBGF("Unknown player: "..dstName);
					end
				else
					self.shouldRemove = false;
				end
			end
		elseif (combatEvent == "SPELL_AURA_APPLIED") then
			if (spellID == 59628) then
				local msg = "";

				if (self.shouldRemove == true) then
					self:DBGF("Attempting to removing the buff.");
					self:RemoveBuff(59628);

					self.shouldRemove = false;

					msg = "[ToTT]: +Damage to ["..self.curTarget.."]";
				else
					msg = "ToTT: +Threat to ["..self.curTarget.."]";
				end

				self:SendTricksMsg(msg);
			end
		end
	end
end

--
-- PLAYER_ROLES_ASSIGNED
--

function SafeTricks:PLAYER_ROLES_ASSIGNED(ev, ...)
	self:ScheduleRosterRescan();
end

--
-- PARTY_MEMBERS_CHANGED
--

function SafeTricks:PARTY_MEMBERS_CHANGED(ev, ...)
	self:ScheduleRosterRescan();
end

--
-- PLAYER_ENTERING_WORLD
--

function SafeTricks:PLAYER_ENTERING_WORLD(ev, ...)
	self:CheckForBG();

	if (self.isInBG == false) then
		self:ScheduleRosterRescan();
	end
end

--
-- Slash command
--

function SafeTricks:SlashHandler(msg)
	self:DBGF("Slash: ["..msg.."]");

	if (msg == "gui") then
		InterfaceOptionsFrame_OpenToCategory(self.optFrame);
	elseif (msg == "disable") then
		if (self:IsEnabled() == true) then
			self:Disable()
			self:Print("Disabled.");
		end
	elseif (msg == "enable") then
		if (self:IsEnabled() == false) then
			self:Enable()
			self:Print("Enabled.");
		end
	elseif (msg == "status") then
		self:PrintStatus();
	elseif (msg == "debug") then
		if (self.debug == true) then
			self.debug = false;
			self:Print("Debugging Disabled.");
		else
			self.debug = true;
			self:Print("Debugging Enabled.");
		end
	elseif (self:IsEnabled() == true) then
		if (msg == "refresh") then
			self:ScheduleRosterRescan();
		elseif (msg == "mute") then
			if (self.isMuted == false) then
				self.isMuted = true;
			else
				self.isMuted = false;
			end
		elseif (msg == "force") then
			self.isForced = true;
			self:Print("Next ToTT will have threat.");
		else
			self:Usage()
		end
	else
		self:Usage()
	end
end

function SafeTricks:Usage()
	self:Print("Safe ToTT: Yawning (Crushridge-US)");
	self:Print("  enable - Enable smart cancel.");
	self:Print("  disable - Disable smart cancel.");
	self:Print("  gui - Display the graphical config.");
	self:Print("  status - Addon status.");

	if (self:IsEnabled() == true) then
		self:Print("  refresh - Rescan the roster for tanks.");
		self:Print("  mute - Mute/Unmute Raid/Party whispers. (NOT SAVED)");
		self:Print("  force - Force the next ToTT to transfer threat. (NOT SAVED)");
	end
end

function SafeTricks:PrintStatus()
	if (self:IsEnabled() == true) then
		local nrTanks = 0;

		if (self.db.char.mode == 0) then
			self:Print("+Threat Tanks, +Damage DPS");
		elseif (self.db.char.mode == 1) then
			self:Print("+Damage Tanks, +Damage DPS");
		elseif (self.db.char.mode == 2) then
			self:Print("+Damage Tanks, +Damage DPS, +Threat Rogues");
		end

		if (self.isInBG == true) then
			self:Print("  Roster scanning is DISABLED. (Player in BG)");
			return;
		end

		for k,v in pairs(self.roster) do
			if ((v.isTank == true) and (v.isAlwaysDPS == false)) then
				self:Print("  ["..nrTanks.."]: "..k.." (Tank)");
				nrTanks = nrTanks + 1;
			elseif (v.isAlwaysTank == true) then
				self:Print("  ["..nrTanks.."]: "..k.." (Forced Tank)");
				nrTanks = nrTanks + 1;
			elseif (v.isAlwaysDPS == true) then
				self:Print("  ["..nrTanks.."]: "..k.." (Forced DPS)");
				nrTanks = nrTanks + 1;
			elseif (v.isRogue == true) then
				if (self.db.char.mode == 2) then
					self:Print("  ["..nrTanks.."]: "..k.." (Rogue)");
					nrTanks = nrTanks + 1;
				end
			end
		end

		self:Print("    "..nrTanks.." players currently tracked.");

		if (self.applyLogic == false) then
			if ((GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0)) then
				self:Print("    * Threat Suppression is DISABLED *");
			end
		end
	else
		self:Print("Addon is Disabled.");
	end
end

--
-- Initalization function.
--

function SafeTricks:OnInitialize()
	-- This addon is relatively (totally) useless if you are not a Rogue.
	if (select(2,UnitClass("player")) == "ROGUE") then
		-- What do you want?
		self:RegisterChatCommand("stricks", "SlashHandler");

		-- Debugging
		self.debug = false;

		-- Options
		self:LoadDefaults();
		self:LoadOptions();

		self.db = LibStub("AceDB-3.0"):New("SafeTricksDB", self.defaults, "char");
		self.db:RegisterDefaults(self.defaults);
		self.optFrame = AceConfigDialog:AddToBlizOptions("SafeTricks", "SafeTricks");
		AceConfig:RegisterOptionsTable("SafeTricks", self.options);

		self:PopulateOptions();

		if (self.db.char.isEnabled == false) then
			self:Disable();
		end
	else
		self:Disable()
	end
end

--
-- User Enable.
--

function SafeTricks:OnEnable()
	self.db.char.isEnabled = true;

	-- Register our event handlers.
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("PLAYER_ROLES_ASSIGNED");
	self:RegisterEvent("PARTY_MEMBERS_CHANGED");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");

	-- Setup our roster timer.
	self.rosterTimer = nil;

	-- Are we actually going to try to cancel buffs?
	self.applyLogic = false;

	-- Set up our default state.
	self.roster = {};
	self.guidRoster = {};
	self.shouldRemove = false;
	self.curTarget = "<Unknown>";
	self.isInBG = false;
	self.isMuted = false;
	self.isForced = false;

	self:CheckForBG();
	self:ScheduleRosterRescan();
end

--
-- User Disable.
--

function SafeTricks:OnDisable()
	self.db.char.isEnabled = false;

	-- Don't need to call our timers anymore.
	self:CancelAllTimers();

	-- Don't need the event handlers anymore.
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:UnregisterEvent("PLAYER_ROLES_ASSIGNED");
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
	self:UnregisterEvent("PLAYER_ENTERING_WORLD");
end

