--
-- SafeTricks_Options.lua: SafeTricks Configuration
-- Yawning (Crushridge-US)
--

--
-- SafeTricks is defined in the main file.
--

if not SafeTricks then return end

--
-- Load the default values into our object.
--

function SafeTricks:LoadDefaults()
	self.defaults = {
		char = {
			isEnabled = true,
			partyAnnounce = false,
			raidAnnounce = false,
			channelAnnounce = false,
			channelSink = "",
			alwaysTank = {},
			alwaysDPS = {},
			mode = 0,
			version = self.version,
		},
	};
end

--
-- Configuration GUI
--

local optionsMode = {
	[0] = "+Threat Tanks, +Damage DPS", -- Default behavior
	[1] = "+Damage Tanks, +Damage DPS", -- Always cancel the buff.
	[2] = "+Threat Tanks, +Damage DPS, +Threat Rogues", -- Rogue Circle-Jerk mode.
};

local overrideEntry = {
	order = 1,
	type = "toggle",
	tristate = true,
	name = function(info) return info[#(info)] end,
	desc = "Checked: Tank, Grey: DPS, Unchecked: Remove",
	set = function(info,val) SafeTricks:OnOverrideSet(info[#(info)], val); end,
	get = function(info) return SafeTricks.optionsCtxt.overrideToggle[info[#(info)]]; end
};

function SafeTricks:LoadOptions()

	-- Announcement related options.
	local announceOptions = {
		name = "Announcements",
		type = "group",
		order = 1,
		args = {
			partyAnnounce = {
				name = "To Party",
				desc = "Announce Tricks to Party",
				type = "toggle",
				order = 0,
				set = function(info, val) SafeTricks.db.char.partyAnnounce = val end,
				get = function(info) return SafeTricks.db.char.partyAnnounce end,
			},
			raidAnnounce = {
				name = "To Raid",
				desc = "Announce Tricks to Raid",
				type = "toggle",
				order = 1,
				set = function(info, val) SafeTricks.db.char.raidAnnounce = val end,
				get = function(info) return SafeTricks.db.char.raidAnnounce end,
			},
			channelAnnounce = {
				name = "To Channel",
				desc = "Announce Tricks to Channel",
				type = "toggle",
				order = 2,
				set = function(info, val) SafeTricks.db.char.channelAnnounce = val end,
				get = function(info) return SafeTricks.db.char.channelAnnounce end,
			},
			channelSink = {
				name = "Channel",
				desc = "Announce Channel",
				type = "input",
				order = 3,
				set = function(info, val) SafeTricks.db.char.channelSink = val end,
				get = function(info) return SafeTricks.db.char.channelSink end,
			},
		},
	};

	-- Override related options.
	local overrideOptions = {
		name = "Overrides",
		type = "group",
		order = 2,
		args = {
			overrides = {
				name = "Current Overrides",
				type = "group",
				inline = true,
				args = {},
				order = 0,
			},
			addOverride = {
				name = "Add Player",
				desc = "Add a player",
				type = "input",
				order = 1,
				set = function(info, val) SafeTricks:AddOverride(val) end,
				get = function(info) end,
			},
		},
	};

	local helpOptions = {
		name = "Help",
		type = "group",
		order = 5,
		args = {
			aboutHeader = {
				name = "About",
				type = "header",
				order = 0,
			},
			about = {
				name = "SafeTricks: Safe Tricks of the Trade - Yawning (Crushridge-US)\n\nIntelligently auto cancel the threat transfer component of Tricks of the Trade while preserving the damage boost.\n\nNote: This mod relies on the LFD group role or the build in Main Tank assignment to autodetect tanks.",
				type = "description",
				order = 2,
			},
			modesHeader = {
				name = "Modes",
				type = "header",
				order = 3,
			},
			modes = {
				name = "Supported modes are:\n* +Threat Tanks, +Damage DPS\n* +Damage Tanks, +Damage DPS\n* +Threat Tanks, +Damage DPS, +Threat Rogues\n\nNote: If no tanks are present (auto-identified/user specified), UNLESS the mode is set to \"+Damage Tanks, +Damage DPS\", the mod will NOT remove the threat transfer.",
				type = "description",
				order = 4,
			},
			overrideHeader = {
				name = "Overrides",
				type = "header",
				order = 5,
			},
			overrides = {
				name = "Use these options to specify players that should always be treated as a tank or DPS.\n\nNote: Role assignment in this manner takes presedence over everyting.  Mode, LFD Role, MT Status, and Class are ALL IGNORED.  Also keep in mind that entries are whitespace and case sensitive.",
				type = "description",
				order = 6,
			},
			commandHeader = {
				name = "Comand Line Interface",
				type = "header",
				order = 7,
			},
			commands = {
				name = "The addon supports the following arguments to \"/stricks\":\n* enable - Enable smart cancel.\n* disable - Disable smart cancel.\n* gui - Display the graphical config.\n* status - Addon status.\n\nAdditionally while the addon is enabled:\n* refresh - Rescan the roster for tanks.\n* mute - Mute/Unmute Raid/Party whispers. (NOT SAVED)\n* force - Force the next ToTT to transfer threat. (NOT SAVED)\n\nCreative use of \"mute\" and \"force\" should allow people to use ToTT in manners similar to what was allowed before installing this addon.",
				type = "description",
				order = 8,
			},
		},
	};

	self.options = {
		name = "SafeTricks".." ("..self.version..")",
		type = "group",
		args = {
			enable = {
				name = "Enable",
				desc = "Enables / Disables the addon",
				type = "toggle",
				order = 0,
				set = function(info, val) SafeTricks:OnEnableSet(val) end,
				get = function(info) return SafeTricks:IsEnabled() end,
			},
			mode = {
				name = "Mode",
				desc = "Configure addon mode",
				type = "select",
				style = "dropdown",
				width = "double",
				order = 1,
				values = optionsMode,
				set = function(info, val) SafeTricks:OnModeSet(val) end,
				get = function(info) return SafeTricks.db.char.mode end,
			},

			announce = announceOptions,
			overrides = overrideOptions,
			help = helpOptions,
		},
	};

	self.optionsCtxt = {
		overrideToggle = {};
	};
end

function SafeTricks:PopulateOptions()
	-- Need to hook OnHide.
	self.optFrame.okay = function() SafeTricks:RebuildOverrideList(); end;
	self.optFrame.cancel = function() SafeTricks:RebuildOverrideList(); end;

	self:RebuildOverrideList();
end

function SafeTricks:FindAlwaysTank(val)
	for i, v in pairs(self.db.char.alwaysTank) do
		if (v == val) then
			return i;
		end
	end

	return nil;
end

function SafeTricks:FindAlwaysDPS(val)
	for i, v in pairs(self.db.char.alwaysDPS) do
		if (v == val) then
			return i;
		end
	end

	return nil;
end


function SafeTricks:OnEnableSet(val)
	if (val == true) then
		self:Enable();
	else
		self:Disable();
	end
end

function SafeTricks:OnModeSet(val)
	self.db.char.mode = val;
	self:DoGUIDRosterUpdate();
end

function SafeTricks:AddOverride(val)
	local idx;

	if (val == "") then return; end

	idx = self:FindAlwaysTank(val);
	if (idx ~= nil) then
		return;
	end

	idx = self:FindAlwaysDPS(val);
	if (idx ~= nil) then
		return;
	end

	self.options.args.overrides.args.overrides.args[val] = overrideEntry; 
	self.optionsCtxt.overrideToggle[val] = false;

	self.options.args.overrides.args.overrides.args.none = nil;
end

function SafeTricks:OnOverrideSet(name, val)
	local idx;
	local rebuild = false;

	self.optionsCtxt.overrideToggle[name] = val;
	
	if (val == true) then
		-- Setting a new tank.
		idx = self:FindAlwaysDPS(name);
		if (idx ~= nil) then
			tremove(self.db.char.alwaysDPS, idx);
		end

		tinsert(self.db.char.alwaysTank, name);
		rebuild = true;
	elseif (val == nil) then
		-- Setting a new DPS.
		idx = self:FindAlwaysTank(name);
		if (idx ~= nil) then
			tremove(self.db.char.alwaysTank, idx);
		end

		tinsert(self.db.char.alwaysDPS, name);
		rebuild = true;
	else
		-- Removing it all together.
		idx = self:FindAlwaysDPS(name);
		if (idx ~= nil) then
			tremove(self.db.char.alwaysDPS, idx);
			rebuild = true;
		end

		idx = self:FindAlwaysTank(name);
		if (idx ~= nil) then
			tremove(self.db.char.alwaysTank, idx);
			rebuild = true;
		end
	end

	if (rebuild == true) then
		--
		-- XXX: self:DoGUIDRosterUpdate() would be sufficient if I
		-- changed the way I update the overrides.
		--
		self:ScheduleRosterRescan();
	end
end

function SafeTricks:RebuildOverrideList()
	local nrPresent = 0;

	table.wipe(self.options.args.overrides.args.overrides.args);
	table.wipe(self.optionsCtxt.overrideToggle);

	for k,v in pairs(self.db.char.alwaysTank) do
		self.options.args.overrides.args.overrides.args[v] = overrideEntry;
		self.optionsCtxt.overrideToggle[v] = true;
		nrPresent = nrPresent + 1;
	end

	for k,v in pairs(self.db.char.alwaysDPS) do
		self.options.args.overrides.args.overrides.args[v] = overrideEntry;
		self.optionsCtxt.overrideToggle[v] = nil;
		nrPresent = nrPresent + 1;
	end

	if (nrPresent == 0) then
		self.options.args.overrides.args.overrides.args.none = {
			name = "You do not have any overrides configured.",
			order = 1,
			type = "description",
		}
	end
end

